#include <Kafka_Provider.h>
#include <memory>
#include <chrono>
#include <unistd.h>
#include <Base_Configurator.h>

using namespace inaf::oasbo::Providers;

KafkaProvider::KafkaProvider() :
		KafkaProvider("127.0.0.1", 9092, "DAQ_topic") {
}

KafkaProvider::KafkaProvider(std::string ip, int port, std::string topic_name) :
		brokerIp(ip), brokerPort(port) {
	setDest(topic_name);
}

void KafkaProvider::pollingThreadFunction() {
	std::cout << "Kafka Provider: polling thread start..." << std::endl;
	while (!this->stopFlag) {
		if (isOpen())
			this->producer->poll(0);
		sleep(1); // pool every second
	}
	std::cout << "Kafka Provider:  polling thread exit..." << std::endl;
}

int KafkaProvider::write(PacketLib::BasePacket &packet) {
	return KafkaProvider::write(packet, this->dest);
}

int KafkaProvider::write(PacketLib::BasePacket &packet, std::string dest) {
	if (!isOpen()) {
		std::cerr << "Kafka provider: provider not open" << std::endl;
		return -1;
	}
	int attempts = 0;
	RdKafka::ErrorCode err;
	while (attempts < 5) { //try to deliver messagr 5 times, then report error.
		err = this->producer->produce(
		/* Topic name */
		dest,
		/* Partition:*/
		0,
		/* Make a copy of the value */
		RdKafka::Producer::RK_MSG_COPY /* Copy payload */,
		/* Value */
		(char*) packet.getBinaryPointer(),
				packet.getHeaderSize() + packet.getPayloadSize()
						+ packet.getTailSize(),
				NULL, 0, 0,
				NULL);
		if (err == RdKafka::ERR_NO_ERROR)
			return 1;

		attempts += 1;

	}
	std::cerr << "Kafka Provider: Failed to produce to topic " << dest << ": "
			<< err << std::endl;
	return -1;
}

int KafkaProvider::open() {
	if (isOpen())
		close();
	RdKafka::Conf *conf = RdKafka::Conf::create(RdKafka::Conf::CONF_GLOBAL);
	std::string errstr;
	if (conf->set("bootstrap.servers",
			std::string(this->brokerIp).append(":").append(
					std::to_string(this->brokerPort)), errstr)
			!= RdKafka::Conf::CONF_OK) {
		std::cerr << errstr << std::endl;
		return -1;
	}
	this->dr_cb = new DeliveryReportCb();
	if (conf->set("dr_cb", dr_cb, errstr) != RdKafka::Conf::CONF_OK) {
		std::cerr << errstr << std::endl;
		return -1;
	}
	this->producer = RdKafka::Producer::create(conf, errstr);
	if (!this->producer) {
		this->producer = nullptr;
		std::cerr << "Kafka Provider: Failed to create producer: " << errstr
				<< std::endl;
		return -1;
	}
	delete conf;
	this->stopFlag = false;
	this->pollThread = std::thread(&KafkaProvider::pollingThreadFunction, this); // start polling
	return 1;
}

int KafkaProvider::close() {
	if (!isOpen())
		return 1;
	std::cout << "Kafka Provider: Flushing kafka queue in 10 seconds.."
			<< std::endl;
	producer->flush(10 * 1000);
	if (producer->outq_len() > 0)
		std::cerr << "Warning: " << producer->outq_len()
				<< " message(s) were not delivered" << std::endl;

	this->stopFlag = true; // Stopping the polling thread
	this->pollThread.join();
	delete producer;
	producer = nullptr;
	delete dr_cb;
	return 1;
}

bool KafkaProvider::isOpen() {
	return this->producer != nullptr;
}

KafkaProvider::~KafkaProvider() {
	close();
}

using K_B = KafkaProvider::Builder;
K_B::Builder() {
	this->reset();
}
K_B::Builder(std::string ip, int port, std::string topic) {
	this->provider = new KafkaProvider(ip, port, topic);

}
K_B::~Builder() {
	delete provider;
}

void K_B::reset() {
	this->provider = new KafkaProvider();
}

K_B* K_B::configFrom(Configurators::BaseConfigurator *conf) {
	std::string target = std::string("KafkaProvider");
	conf->readConfigFromSource(target);
	std::map<std::string, std::string> params = conf->getConfig();

	std::string key = target+"_"+ip_key;
	if (params.count(key) > 0)
		provider->brokerIp = params[key];

	key = target+"_"+port_key;
	if (params.count(key) > 0)
		provider->brokerPort = std::stoi(params[key]);

	key = target+"_"+topic_key;
	if (params.count(key) > 0)
		provider->setDest(params[key]);
	return this;
}

K_B* K_B::setIp(std::string ip) {
	provider->brokerIp = ip;
	return this;
}

K_B* K_B::setPort(int port) {
	provider->brokerPort = port;
	return this;
}

K_B* K_B::setTopic(std::string topic) {
	provider->setDest(topic);
	return this;
}

KafkaProvider* K_B::getProvider() {
	KafkaProvider *result = this->provider;
	this->reset();
	return result;
}
